#---------------------------------------------------------------------
#   Mersenne Twister
#   file : mt19937.s  121steps 484bytes
#     Rewritten in PowerPC Assembly by Jun Mizutani 2005/04/05.
#     From original code in C by Takuji Nishimura(mt19937int.c).
#     PPC version Copyright (C) 2005 Jun Mizutani.
#---------------------------------------------------------------------

# A C-program for MT19937: Integer version (1999/10/28)
#  genrand() generates one pseudorandom unsigned integer (32bit)
# which is uniformly distributed among 0 to 2^32-1  for each
# call. sgenrand(seed) sets initial values to the working area
# of 624 words. Before genrand(), sgenrand(seed) must be
# called once. (seed is any 32-bit integer.)
#   Coded by Takuji Nishimura, considering the suggestions by
# Topher Cooper and Marc Rieffel in July-Aug. 1997.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later
# version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Library General Public License for more details.
# You should have received a copy of the GNU Library General
# Public License along with this library; if not, write to the
# Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307  USA
#
# Copyright (C) 1997, 1999 Makoto Matsumoto and Takuji Nishimura.
# Any feedback is very welcome. For any question, comments,
# see http://www.math.keio.ac.jp/matumoto/emt.html or email
# matumoto#math.keio.ac.jp
#
# REFERENCE
# M. Matsumoto and T. Nishimura,
# "Mersenne Twister: A 623-Dimensionally Equidistributed Uniform
# Pseudo-Random Number Generator",
# ACM Transactions on Modeling and Computer Simulation,
# Vol. 8, No. 1, January 1998, pp 3--30.

#---------------------------------------------------------------
# load 32bit immediate
#---------------------------------------------------------------
.ifndef __L32I
     __L32I = 1
    .macro  L32I    reg, symbol
        lis     \reg, \symbol@ha
        addi    \reg, \reg, \symbol@l
    .endm
.endif

#---------------------------------------------------------------
.section ".text"
.align 2

    .equ    N,      624
    .equ    M,      397
    .equ    ffff0000,           0xffff0000
    .equ    n69069,             69069
    .equ    TEMPERING_MASK_B,   0x9d2c5680
    .equ    TEMPERING_MASK_C,   0xefc60000
    .equ    UPPER_MASK,         0x80000000
    .equ    LOWER_MASK,         0x7fffffff
    .equ    MATRIX_A,           0x9908b0df

#---------------------------------------------------------------
# Initialize Mersenne Twister
#   enter r3 : seed
#---------------------------------------------------------------
sgenrand:
        subi    r1, r1, 28
        stswi   r3, r1, 28      # push r3-r9

        L32I    r6, mt
        li      r7, N
        L32I    r4, mti
        stw     r7, 0(r4)       # mti=N
        mtctr   r7
        L32I    r8, ffff0000
        L32I    r9, n69069
        li      r7, 0           # I=0
    1:
        and     r4, r3, r8      # n=seed & 0xffff0000
        mullw   r3, r9, r3      # S=seed*69069
        addi    r3, r3, 1       # S=seed*69069 + 1
        and     r5, r3, r8      # m=S & 0xffff0000
        srwi    r5, r5, 16      # m=(S&0xffff0000)>>16
        or      r4, r4, r5      # m=n|seed & 0xffff0000
        mullw   r3, r9, r3
        addi    r3, r3, 1
        stwx    r4, r6, r7      # mt[I]=S|seed & 0xffff0000
        addi    r7, r7, 4       # I=I+4
        bdnz    1b

        lswi    r3, r1, 28      # pop r3-r9
        addi    r1, r1, 28      # restore r1
        blr

#---------------------------------------------------------------
# Generate Random Number
#   return r0 : random number
#---------------------------------------------------------------
genrand:
        subi    r1, r1, 32
        stswi   r3, r1, 0       # push r3-r10
        subi    r1, r1, 16
        stswi   r11, r1, 16     # push r11-r14
        mflr    r14

        L32I    r5, mti
        L32I    r6, mt
        li      r7, N
        lwz     r0, 0(r5)       # mti
        cmpw    r0, r7          # 623
        blt     3f              # from mt[]
        li      r8, M
        L32I    r9, UPPER_MASK
        L32I    r10, LOWER_MASK
        L32I    r11, MATRIX_A
        li      r4, 0           # K=0
        slwi    r12, r4, 2      # r12= offset K
    1:
        bl      rnd_common0     # return Y>>1:r0,Z:r1
        add     r13, r4, r8     # r13 = K + M
        bl      rnd_common
        stwx    r3, r6, r12     # mt[K]=P^Q^Z
        addi    r4, r4, 1       # K=K+1
        slwi    r12, r4, 2      # r12=4K
        sub     r0, r7, r8      # N-M
        cmpw    r4, r0
        blt     1b
    2:
        bl      rnd_common0     # return Y>>1:r0,Z:r1
        sub     r13, r8, r7     # r13 = M-N
        add     r13, r4, r13    # K+(M-N)
        bl      rnd_common
        stwx    r3, r6, r12     # mt[K]=P^Q^Z
        addi    r4, r4, 1       # K=K+1
        slwi    r12, r4, 2      # r12=4K
        subi    r13, r7, 1      # 623
        cmpw    r4, r13
        blt     2b
        li      r3, 0           # J=0
        bl      rnd_common1     # return Y>>1:r0,Z:r1
        subi    r13, r8, 1      # 396
        bl      rnd_common
        subi    r13, r7, 1      # 623
        slwi    r12, r13, 2
        stwx    r3, r6, r12     # mt[623]=P^Q^Z
        li      r0, 0
        stw     r0, 0(r5)       # mti=0
    3:
        L32I    r9, TEMPERING_MASK_B
        L32I    r10, TEMPERING_MASK_C
        lwz     r4, 0(r5)       # mti
        slwi    r13, r4, 2
        lwzx    r3, r6, r13     # y=mt[mti]
        addi    r4, r4, 1
        stw     r4, 0(r5)       # mti++
        srwi    r0, r3, 11      # y>>11
        xor     r3, r3, r0      # y=y^(y>>11)
        slwi    r0, r3, 7       # y << 7
        and     r0, r0, r9      # TEMPERING_MASK_B
        xor     r3, r3, r0
        slwi    r0, r3, 15
        and     r0, r0, r10     # TEMPERING_MASK_C
        xor     r3, r3, r0
        srwi    r0, r3, 18
        xor     r0, r3, r0

        mtlr    r14             # restore lr
        lswi    r11, r1, 16     # pop r11-r13
        addi    r1, r1, 16      # restore r1
        lswi    r3, r1, 32     # pop r3-r10
        addi    r1, r1, 32      # restore r1
        blr

    rnd_common0:
        addi    r3, r4, 1       # J=K+1
    rnd_common1:
        lwzx    r0, r6, r12     # r0 = mt[K]  (mt+K*4)
        and     r0, r0, r9      # r0 = mt[K] & UPPER_MASK
        slwi    r13, r3, 2      # r13 = offset J
        lwzx    r13, r6, r13    # r1 = mt[J]
        and     r13, r13, r10   # r1 = r1 & LOWER_MASK
        or      r13, r0, r13    # r13=y
        srwi    r0, r13, 1      # r0=(y>>1)
        li      r3, 0
        andi.   r13, r13, 1     # if (y&1) r1=0
        beqlr
        mr      r3, r11         # else r1 = MATRIX_A
        blr

    rnd_common:
        slwi    r13, r13, 2
        lwzx    r13, r6, r13    # mt[x]
        xor     r13, r13, r0    # mt[x]^P
        xor     r3, r13, r3     # mt[x]^P^Z
        blr

#---------------------------------------------------------------
.section ".bss"
.align 2
mti:                .long   N + 1
mt:                 .skip   624 * 4
