# ------------------------------------------------------------------------
# Standard I/O Subroutine for PowerPC
#   2004/08/01-12, 2005/03/16,18,19,4/6,11
#
# as -mregnames -o $1.o $1.s
#
# Copyright (C) 2004  Jun Mizutani <mizutani.jun @ nifty.ne.jp>
# stdio.s may be copied under the terms of the GNU General Public License.
# ------------------------------------------------------------------------

.ifndef __STDIO
__STDIO = 1

.ifndef __SYSCALL
  .equ sys_exit,  1
  .equ sys_read,  3
  .equ sys_write, 4
.endif

.text

#------------------------------------
# exit with 0
# long sys_exit(int error_code)
Exit:
        li      r3, 0
        li      r0, sys_exit
        sc
        blr

#------------------------------------
# exit with r3
# long sys_exit(int error_code)
ExitN:
        li      r0, sys_exit
        sc
        blr

#------------------------------------
# print string to stdout
# input   r3 : address, r4 : length
# ssize_t sys_write(unsigned int fd, const char * buf, size_t count)
OutString:
        stwu    r0, -4(r1)      # push r0
        stwu    r5, -4(r1)      # push r5
        mr      r5, r4
        mr      r4, r3
        li      r3, 1           # r3  stdout
        li      r0, sys_write
        sc
        mr      r3, r4
        mr      r4, r5
        lwz     r5, 0(r1)       # pop r5
        lwzu    r0, 4(r1)       # pop r0
        addi    r1, r1, 4       # restore r1
        blr                     # return


#------------------------------------
# input   r3 : address
# output  r0 : length of strings
StrLen:
        subi    r1, r1, 12
        stswi   r4, r1, 12      # push r4-r6
        #mfcr    r6
        addi    r5, r3, -1      # initialize
    1:
        lbzu    r4, 1(r5)       # read s[i] and i++
        cmpwi   r4, 0           # compare s[i] to e\0f
        bne     1b              # if (s[i] != e\0f) goto loop
        subf    r0, r3, r5      # r0=r5-r3
        #mtcr    r6
        lswi    r4, r1, 12      # pop r4-r6
        addi    r1, r1, 12      # restore r1
        blr

#------------------------------------
# print asciiz string
# r3 : address of string
OutAsciiZ:
        subi    r1, r1, 12
        stswi   r4, r1, 12      # push r4-r6
        mr      r6, r0
        mflr    r5
        bl      StrLen
        mr      r4, r0
        bl      OutString
        mtlr    r5              # restore lr
        mr      r0, r6
        lswi    r4, r1, 12      # pop r4-r6
        addi    r1, r1, 12      # restore r1
        blr

#------------------------------------
# print pascal string to stdout
# r3 : top address
OutPString:
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # push lr
        stwu    r4, -4(r1)
        lbz     r4, 0(r3)
        addi    r3, r3, 1
        bl      OutString
        subi    r3, r3, 1       # restore r3
        lwz     r4, 0(r1)       # pop lr
        mtlr    r4              # restore lr
        lwzu    r4, 4(r1)       # pop r4
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# print 1 character to stdout
# r3 : put char
OutChar:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mr      r6, r0          # save r0
        addi    r4, r1, 3       # r4  address (for Big Endian)
        li      r3, 1           # r3  stdout
        mr      r5, r3          # r5  length
        li      r0, sys_write
        sc
        mr      r0, r6          # restore r0
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#------------------------------------
# print 4 characters in r3 to stdout
OutChar4:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mflr    r6              # save lr
        mr      r4, r3
        li      r5, 4
1:
        rlwinm  r4, r4, 8, 0, 31 # rotate left 8bit
        andi.   r3, r4, 0x7F
        cmpwi   r3, 0x7F
        beq     2f
        cmpwi   r3, 0x20
        bge     3f
2:      li      r3, '.'
3:
        bl      OutChar
        subic.  r5, r5, 1
        bgt     1b
        mtlr    r6              # restore lr
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#------------------------------------
# new line
NewLine:
        stwu    r3, -4(r1)      # push r3
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # save lr
        li      r3, 10
        bl      OutChar
        mtlr    r4              # restore lr
        lwz     r4, 0(r1)       # pop r4
        lwzu    r3, 4(r1)       # pop r3
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# Backspace
BackSpace:
        stwu    r3, -4(r1)      # push r3
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # save lr
        li      r3, 8
        bl      OutChar
        li      r3, ' '
        bl      OutChar
        li      r3, 8
        bl      OutChar
        mtlr    r4              # restore lr
        lwz     r4, 0(r1)       # pop r4
        lwzu    r3, 4(r1)       # pop r3
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# print binary number
#   r3 : number
#   r4 : bit
PrintBinary:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mflr    r6              # save lr
        cmpwi   r4, 0           # r4 > 0 ?
        beq     4f              # if r4=0 exit

        mr      r5, r3
        li      r3, 32
        cmpw    r4, r3
        ble     1f
        mr      r4, r3          # if r4>32 then r4=32
    1:  sub     r3, r3, r4
        slw     r5, r5, r3      # skip 32-r4 bits
    2:  li      r3, '0'
        or.     r5, r5, r5
        bge     3f
        addi    r3, r3, 1
    3:  bl      OutChar
        slwi    r5, r5, 1       # next bit
        subic.  r4, r4, 1
        bne     2b
    4:
        mtlr    r6              # restore lr
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#------------------------------------
# print r4 digit octal number
#   r3 : number
#   r4 : columns
PrintOctal:
        subi    r1, r1, 12
        stswi   r5, r1, 12      # push r5-r7
        mflr    r7              # save lr

        cmpwi   r4, 0           # r4 > 0 ?
        beq     4f              # if r4=0 exit
        mr      r6, r4          # column
    1:  andi.   r5, r3, 7
        srwi    r3, r3, 3
        stwu    r5, -4(r1)      # push lower digit
        subic.  r6, r6, 1       # column--
        bne     1b
    2:
        lwz     r3, 0(r1)       # pop upper digit
        addi    r1, r1, 4       # restore r1
        addi    r3, r3, '0'     # convert to ascii
        bl      OutChar
        subic.  r4, r4, 1       # column--
        bne     2b

    4:  mtlr    r7              # restore lr
        lswi    r5, r1, 12      # pop r5-r7
        addi    r1, r1, 12      # restore r1
        blr

#------------------------------------
# print 2 digit hex number (lower 8 bit of r3)
#   r3 : number
PrintHex2:
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # (1) push lr
        stwu    r4, -4(r1)      # (2) push lr
        li      r4, 2
        bl      PrintHex
        lwz     r4, 0(r1)       # (1) pop lr
        mtlr    r4              # (2) pop lr
        lwzu    r4, 4(r1)       # pop r4
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# print 4 digit hex number (lower 16 bit of r3)
#   r3 : number
PrintHex4:
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # (1) push lr
        stwu    r4, -4(r1)      # (2) push lr
        li      r4, 4
        bl      PrintHex
        lwz     r4, 0(r1)       # (1) pop lr
        mtlr    r4              # (2) pop lr
        lwzu    r4, 4(r1)       # pop r4
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# print 8 digit hex number (r3)
#   r3 : number
PrintHex8:
        stwu    r4, -4(r1)      # push r4
        mflr    r4              # (1) push lr
        stwu    r4, -4(r1)      # (2) push lr
        li      r4, 8
        bl      PrintHex
        lwz     r4, 0(r1)       # (1) pop lr
        mtlr    r4              # (2) pop lr
        lwzu    r4, 4(r1)       # pop r4
        addi    r1, r1, 4       # restore r1
        blr

#------------------------------------
# print hex number
#   r3 : number     r4 : digit
PrintHex:
        subi    r1, r1, 20
        stswi   r3, r1, 20      # push r3-r7
        mflr    r7              # save lr

        mr      r6, r4          # column
    1:  andi.   r5, r3, 0x0F    #
        srwi    r3, r3, 4       #
        ori     r5, r5, 0x30
        cmpwi   r5, 0x39
        ble     2f
        addi    r5, r5, '@'-'9' # if (r5>'9') r5+='@'-'9'
    2:
        stwu    r5, -4(r1)      # push least digit
        subic.  r6, r6, 1       # column--
        bne     1b
        mr      r6, r4          # column
    3:
        lwz     r3, 0(r1)       # pop most digit
        addi    r1, r1, 4       # restore r1
        bl      OutChar
        subic.  r6, r6, 1       # column--
        bne     3b

    4:  mtlr    r7              # restore lr
        lswi    r3, r1, 20      # pop r3-r7
        addi    r1, r1, 20      # restore r1
        blr

#------------------------------------
# Output Unsigned Number to stdout
# r3 : number
PrintLeftU:
        subi    r1, r1, 28
        stswi   r3, r1, 28      # push r3-r9
        mflr    r9              # save lr
        li      r5, 0           # counter
        li      r4, 0           # positive flag
        b       1f

#------------------------------------
# Output Number to stdout
# r3 : number
PrintLeft:
        subi    r1, r1, 28
        stswi   r3, r1, 28      # push r3-r9
        mflr    r9              # save lr

        li      r5, 0           # counter
        li      r4, 0           # positive flag
        cmpwi   r3, 0
        bge     1f
        li      r4, 1           # set negative
        subfic  r3, r3, 0       # r3 = 0-r3

    1:  mr      r6, r3
        li      r7, 10          # r7 = 10
        divwu   r3, r3, r7      # r3 = r3/10
        mullw   r8, r3, r7      # r8 = quotient * 10
        subf    r8, r8, r6      # r8 = mod(r3/10)

        addi    r5, r5, 1       # counter++
        stwu    r8, -4(r1)      # least digit (reminder)
        cmpwi   r3, 0
        bne     1b              # done ?

        cmpwi   r4, 0
        beq     2f
        li      r3, '-'         # if (r3<0) putchar("-")
        bl      OutChar         # output '-'
    2:
        lwz     r3, 0(r1)       # pop  most digit
        addi    r1, r1, 4       # restore r1
        addi    r3, r3, '0'     # ASCII
        bl      OutChar         # output a digit
        subic.  r5, r5, 1       # counter--
        bne     2b

        mtlr    r9              # restore lr
        lswi    r3, r1, 28      # pop r3-r9
        addi    r1, r1, 28      # restore r1
        blr

#------------------------------------
# Output Number to stdout
# r3:column
# r4:number
PrintRight0:
        subi    r1, r1, 32
        stswi   r3, r1, 32      # push r3-r10
        mflr    r5              # save lr
        stwu    r5, -4(r1)
        li      r7, '0'
        b       0f

#------------------------------------
# Output Unsigned Number to stdout
# r3:column
# r4:number
PrintRightU:
        subi    r1, r1, 32
        stswi   r3, r1, 32      # push r3-r10
        mflr    r5              # save lr
        stwu    r5, -4(r1)
        li      r7, ' '
    0:  mr      r8, r4
        li      r5, 0           # counter
        li      r4, 0           # positive flag
        b       1f              # PrintRight.1

#------------------------------------
# Output Number to stdout
# r3:number
# r4:column
PrintRight:
        subi    r1, r1, 32
        stswi   r3, r1, 32      # push r3-r10
        mflr    r5              # save lr
        stwu    r5, -4(r1)

        li      r7, ' '
        mr      r8, r4          # colum
        li      r5, 0           # counter
        li      r4, 0           # positive flag
        cmpwi   r3, 0           # check negative
        bge     1f
        li      r4, 1           # set negative
        subfic  r3, r3, 0       # r3 = 0-r3
    1:  mr      r6, r3
        li      r9, 10          # r9 = 10
        divwu   r3, r3, r9      # r3 = r3/10
        mullw   r10, r3, r9     # r10 = quotient * 10
        subf    r10, r10, r6    # r10 = mod(r3/10)
        addi    r5, r5, 1       # counter++
        stwu    r10, -4(r1)     # least digit (reminder)
        cmpwi   r3, 0
        bne     1b              # done ?

        sub.    r8, r8, r5      # r8 = no. of space
        ble     3f              # dont write space
        cmpwi   r4, 0           # check sign
        beq     2f
        subic   r8, r8, 1       # reserve space for -
    2:  mr      r3, r7          # output space or '0'
        bl      OutChar
        subic.  r8, r8, 1       # nspace--
        bgt     2b

    3:  cmpwi   r4, 0
        beq     4f
        li      r3, '-'         # if (r3<0) putchar("-")
        bl      OutChar         # output '-'
    4:
        lwz     r3, 0(r1)       # pop  most digit
        addi    r1, r1, 4       # restore r1
        addi    r3, r3, '0'     # ASCII
        bl      OutChar         # output a digit
        subic.  r5, r5, 1       # counter--
        bne     4b

        lwz     r5, 0(r1)
        addi    r1, r1, 4       # restore r1
        mtlr    r5              # restore lr
        lswi    r3, r1, 32      # pop r3-r10
        addi    r1, r1, 32      # restore r1
        blr

#------------------------------------
# input 1 character from stdin
# r0 : get char
InChar:
        stwu    r5, -4(r1)      # push r5
        stwu    r4, -4(r1)      # push r4
        stwu    r3, -4(r1)      # push r3
        li      r0, 0           # clear buffer
        stwu    r0, -4(r1)      # push r0, buffer
        addi    r4, r1, 3       # r4  address (for Big Endian)
        li      r3, 0           # r3  stdin
        li      r5, 1           # r5  length
        li      r0, sys_read
        sc
        lwz     r3, 4(r1)       # pop r3
        lwz     r4, 8(r1)       # pop r4
        lwz     r5, 12(r1)       # pop r5
        lwz     r0, 0(r1)       # pop r0
        addi    r1, r1, 16       # restore r1
        blr

#------------------------------------
# Input Line
# r3 : BufferSize
# r4 : Buffer Address
# return       r0 : no. of char
InputLine0:
        subi    r1, r1, 24
        stswi   r3, r1, 24      # push r3-r8
        mflr    r8              # save lr

        mr      r6, r3          # BufferSize
        mr      r7, r4          # Input Buffer
        li      r5, 0           # counter
    1:
        bl      InChar
        mr      r4, r0
        cmpwi   r4, 0x08        # BS ?
        bne     2f
        cmpwi   r5, 0
        beq     2f
        bl      BackSpace       # backspace
        subic   r5, r5, 1
        b       1b
    2:
        cmpwi   r4, 0x0A        # enter ?
        beq     4f              # exit
        bl      OutChar         #
        stbx    r4, r7, r5      # store a char into buffer
        addi    r5, r5, 1
        cmpw    r5, r6
        blt     1b
    3:
        subic   r5, r5, 1
        bl      BackSpace
        b       1b

    4:  li      r3, 0
        stbx    r3, r7, r5      # store end of line (0)
        addi    r5, r5, 1
        bl      NewLine
        mr      r0, r5
        mtlr    r8              # restore lr
        lswi    r3, r1, 24      # pop r3-r8
        addi    r1, r1, 24      # restore r1
        blr

.endif
