#-------------------------------------------------------------------------
# file : termio.s
# 2005/04/11,05/05,08,12
# Copyright (C) 2005 Jun Mizutani <mizutani.jun@nifty.ne.jp>
# termio.s may be copied under the terms of the GNU General Public License.
#-------------------------------------------------------------------------

.ifndef __TERMIO
__TERMIO = 1

.include "syscalls.s"

#TIOCGWINSZ   = 0x5413
TIOCSWINSZ   = 0x80087467
TIOCGWINSZ   = 0x40087468

NCCS  = 19

#  c_cc characters
VINTR       = 0
VQUIT       = 1
VERASE      = 2
VKILL       = 3
VEOF        = 4
VMIN        = 5
VEOL        = 6
VTIME       = 7
VEOL2       = 8
VSWTC       = 9
VWERASE     = 10
VREPRINT    = 11
VSUSP       = 12
VSTART      = 13
VSTOP       = 14
VLNEXT      = 15
VDISCARD    = 16

#  c_lflag bits
ISIG        = 0x00000080
ICANON      = 0x00000100
XCASE       = 0x00004000
ECHO        = 0x00000008
ECHOE       = 0x00000002
ECHOK       = 0x00000004
ECHONL      = 0x00000010
NOFLSH      = 0x80000000
TOSTOP      = 0x00400000
ECHOCTL     = 0x00000040
ECHOPRT     = 0x00000020
ECHOKE      = 0x00000001
FLUSHO      = 0x00800000
PENDIN      = 0x20000000
IEXTEN      = 0x00000400

TCGETS    = 0x402C7413
TCSETS    = 0x802C7414

SEEK_SET  = 0               # Seek from beginning of file.
SEEK_CUR  = 1               # Seek from current position.
SEEK_END  = 2               # Seek from end of file.

termios_mode = ~ICANON & ~ECHO & ~ECHONL

#---------------------------------------------------------------
# load 32bit immediate
#---------------------------------------------------------------
.ifndef __LWI
     __LWI = 1
    .macro  LWI    reg, symbol
        lis     \reg, \symbol@ha
        addi    \reg, \reg, \symbol@l
    .endm
.endif


.text
#--------------------------------------------------------------
# ߤ termios ¸
#--------------------------------------------------------------
GET_TERMIOS:
        subi    r1, r1, 24
        stswi   r3, r1, 24      # push r3-r8
        mflr    r8
        LWI     r5, old_termios
        bl      tcgetattr
        LWI     r4, new_termios
        sub     r3, r4, r5
        li      r7, 0
    1:
        lwzx    r6, r5, r7
        stwx    r6, r4, r7
        addi    r7, r7, 4
        cmpw    r7, r3
        bne     1b
        mtlr    r8
        lswi    r3, r1, 24      # pop r3-r8
        addi    r1, r1, 24      # restore r1
        blr

#--------------------------------------------------------------
#  termios 
# Raw⡼, ECHO ̵, ECHONL ̵
# VTIME=0, VMIN=1 : 1Хɤ߼ޤԵ
#--------------------------------------------------------------
SET_TERMIOS:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mflr    r6
        LWI     r3, new_termios
        lwz     r4, 12(r3)      # c_lflag
        LWI     r5, termios_mode
        and     r4, r4, r5
        ori     r4, r4, ISIG
        stw     r4, 12(r3)
        LWI     r3, nt_c_cc
        li      r4, 1
        stb     r4, VMIN(r3)
        li      r4, 0
        stb     r4, VTIME(r3)
        LWI     r5, new_termios
        bl      tcsetattr
        mtlr    r6
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#--------------------------------------------------------------
# ¸Ƥ termios 
#--------------------------------------------------------------
RESTORE_TERMIOS:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mflr    r6
        LWI     r5, old_termios
        bl      tcsetattr
        mtlr    r6
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#--------------------------------------------------------------
# ɸϤ termios μ
# tcgetattr(&termios)
# tcsetattr(&termios)
# r5 : buffer adress
# r4 : destroy
#--------------------------------------------------------------
tcgetattr:
        LWI     r4, TCGETS
        b       IOCTL

tcsetattr:
        LWI     r4, TCSETS

#--------------------------------------------------------------
# ɸϤ ioctl μ¹
# sys_ioctl(unsigned int fd, unsigned int cmd,
#           unsigned long arg)
# r4 : cmd
# r5 : buffer adress
#--------------------------------------------------------------
IOCTL:
        stwu    r0, -4(r1)      # push r0
        stwu    r3, -4(r1)      # push r3
        li      r3, 0           # 0 : to stdin
        li      r0, sys_ioctl
        sc
        lwz     r3, 0(r1)       # pop r3
        lwzu    r0, 4(r1)       # pop r0
        addi    r1, r1, 4       # restore r1
        blr                     # return

#-------------------------------------------------------------------------
# get window size
# r0 : column(upper 16bit), raw(lower 16bit)
#-------------------------------------------------------------------------
WinSize:
        subi    r1, r1, 12
        stswi   r3, r1, 12      # push r3-r5
        li      r3, 0           # to stdout
        LWI     r4, TIOCGWINSZ  # get wondow size
        LWI     r5, winsize
        li      r0, sys_ioctl
        sc
        lwz     r0, 0(r5)       # winsize.ws_row
        rlwinm  r0, r0, 16, 0, 31 # exchange 16bit
        lswi    r3, r1, 12      # pop r3-r5
        addi    r1, r1, 12      # restore r1
        blr

#--------------------------------------------------------------
# input 1 character from stdin
# r0 : get char (0:not pressed)
#--------------------------------------------------------------
RealKey:
        subi    r1, r1, 16
        stswi   r3, r1, 16      # push r3-r6
        mflr    r6

        LWI     r3, nt_c_cc
        li      r4, 0
        stb     r4, VMIN(r3)
        LWI     r5, new_termios
        bl      tcsetattr

        li      r0, 0           # clear buffer
        stwu    r0, -4(r1)      # push r0, buffer
        addi    r4, r1, 3       # r4  address (for Big Endian)
        li      r3, 0           # r3  stdin
        li      r5, 1           # r5  length
        li      r0, sys_read
        sc
        lwz     r0, 0(r1)       # pop r0
        addi    r1, r1, 4       # restore r1
        LWI     r3, nt_c_cc
        li      r4, 1
        stb     r4, VMIN(r3)
        LWI     r5, new_termios
        bl      tcsetattr
        mtlr    r6
        lswi    r3, r1, 16      # pop r3-r6
        addi    r1, r1, 16      # restore r1
        blr

#==============================================================
.section    ".bss"
                    .align  2
old_termios:
ot_c_iflag:         .long   1           # input mode flags
ot_c_oflag:         .long   1           # output mode flags
ot_c_cflag:         .long   1           # control mode flags
ot_c_lflag:         .long   1           # local mode flags
ot_c_cc:            .skip   NCCS        # control characters
ot_c_line:          .byte   1           # line discripline
ot_c_ispeed:        .long   1           # input speed
ot_c_ospeed:        .long   1           # output speed

                    .align 2
new_termios:
nt_c_iflag:         .long   1           # input mode flags
nt_c_oflag:         .long   1           # output mode flags
nt_c_cflag:         .long   1           # control mode flags
nt_c_lflag:         .long   1           # local mode flags
nt_c_cc:            .skip   NCCS        # control characters
nt_c_line:          .byte   1           # line discripline
nt_c_ispeed:        .long   1           # input speed
nt_c_ospeed:        .long   1           # output speed

winsize:
ws_row:             .hword  1
ws_col:             .hword  1
ws_xpixel:          .hword  1
ws_ypixel:          .hword  1

ru:                                 # 18 words
ru_utime_tv_sec:    .long   1       # user time used
ru_utime_tv_usec:   .long   1       #
ru_stime_tv_sec:    .long   1       # system time used
ru_stime_tv_usec:   .long   1       #
ru_maxrss:          .long   1       # maximum resident set size
ru_ixrss:           .long   1       # integral shared memory size
ru_idrss:           .long   1       # integral unshared data size
ru_isrss:           .long   1       # integral unshared stack size
ru_minflt:          .long   1       # page reclaims
ru_majflt:          .long   1       # page faults
ru_nswap:           .long   1       # swaps
ru_inblock:         .long   1       # block input operations
ru_oublock:         .long   1       # block output operations
ru_msgsnd:          .long   1       # messages sent
ru_msgrcv:          .long   1       # messages received
ru_nsignals:        .long   1       # signals received
ru_nvcsw:           .long   1       # voluntary context switches
ru_nivcsw:          .long   1       # involuntary

.endif
